/* SPDX-License-Identifier: BSD-3-Clause
 *
 * Copyright 2023 The ChromiumOS Authors
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

package gvip_aximm_pkg;
  import gvip_policy_pkg::policy;
  import greporter_pkg::greporter;
  import gvip_handshake_pkg::*;
  import gaximm_pkg::*;
  import gvip_aximm_types_pkg::gvip;


  //////////////////////////////
  // Analysis ports interface //
  //////////////////////////////

  // Interface Class: analysis_port_ic
  //
  // Subscriber should implement this interface
  // Object with this interface can be registered as Anaysis Port in VIP Core
  // Depending on registration type write method will be called after each transfer
  // (core::add_ap2transfer) or after whole packet (core::add_ap2pkt)
  //
  interface class analysis_port_ic #(type item_t = logic);
    pure virtual function item_t write (input item_t item);
  endclass : analysis_port_ic


  ///////////////////////
  // VIP configuration //
  ///////////////////////


  // Class: gvip_config
  //
  // Configuration object to configure VIP - can be pass to constructor
  //
  class gvip_config;
    // Variable: stop_vip
    //
    // If this is set to 1, VIP will not resume main thread after 'clear'
    // all VIPs thread will be stopped
    //
    bit stop_vip = 0;

    // Variable: transaction_in_reset
    //
    // In normal operation VIP operates only if reset in gclk_if is in not active state
    // If this variable is set, VIP does not take reset state into account
    //
    bit transaction_in_reset = 0;

    // Variable: report_id
    //
    // This string will be used as ID of messages generated by VIP
    //
    string report_id = "";
  endclass : gvip_config


  /////////////////////////////
  // Internal implementation //
  /////////////////////////////

  `include "gvip_aximm_imp.svh"


  /////////
  // VIP //
  /////////


  // Class: core
  //
  // VIP core class
  // parameters must be consistent with interface insance parameters
  //
  // Parameters:
  //   N      - number of bytes of tdata
  //   I      - id signal width
  //   A      - address signal width
  //   U_REQ  - awuser and aruser type
  //   U_DATA - wuser and ruser type
  //   U_RESP - buser type
  //   item_t - item type definition that matches interface -
  //            for some simulators is requied to have single point of
  //            type definition so in such case type can by pass to VIP
  class core #(
                int N=1,
                int I=1,
                int A=1,
                type U_REQ=logic[0:0],
                type U_DATA=logic[0:0],
                type U_RESP=logic[0:0],
                type item_t = gvip_aximm_item_pkg::item #(N, I, A, U_REQ, U_DATA, U_RESP)
              ) extends core_imp #(N, I, A, U_REQ, U_DATA, U_RESP, item_t);

    // Function: new
    //
    // Creates an instance of VIP core
    //
    // name     - instance name
    // vif      - AXI Memory Mapped interface handler
    // vclk     - clock interface handler
    // vip_type - one of VIPs type (MASTER/SLAVE/MONITOR)
    // vip_cfg  - optional cofiguration object (please check: gvip_config)
    //
    extern function new (string name,
                         virtual gaximm_if #(N, I, A, U_REQ, U_DATA, U_RESP) vif,
                         virtual gclk_if vclk,
                         gvip::type_e vip_type,
                         gvip_config vip_cfg = null);

    // Function: get_vip_type
    //
    // Returns type of VIP
    //
    extern virtual function gvip::type_e get_vip_type();

    // Function: get_config
    //
    // Returns current configuration object
    //
    extern virtual function gvip_config get_config();

    // Function: add_ap2pkt
    //
    // Use this function to register object with analysis port interface implementation,
    // ~write~ function from this interface will be called with full package
    // Any number of object can be register but the same object instance
    // will be given to all subscribers
    //
    extern virtual function void add_ap2pkt(gvip::channel_e channel,
                                            analysis_port_ic #(item_t) ap);

    // Function: add_ap2transfer
    //
    // Use this function to register object with analysis port interface implementation,
    // ~write~ function from this interface will be called with each transfer occurrence
    // Any number of object can be register but the same object instance
    // will be given to all subscribers
    //
    extern virtual function void add_ap2transfer(gvip::channel_e channel,
                                                 analysis_port_ic #(item_t) ap);

    // Function: set_policy
    //
    // Assign new handshake policy to VIP's channel
    //
    // channel - one of channels (AW/AR/W/R/B/AW_W)
    //           In case of AW_W the same policy will be apply to AW and AR channel
    // obj - policy object derived from gvip_policy_pkg::policy
    // do_randomize_policy - if set, VIP will call randomize method at policy object,
    //                       usefull when policy is created without declaration e.g.:
    //                       vip.set_policy(gvip_policy_pkg::random::create("def_random"), 1);
    // clone - in default VIP will clone policy to operate on it
    //
    extern virtual function void set_policy(gvip::channel_e channel, policy obj,
                                            bit do_randomize_policy = 0, bit clone = 1);

    // Function: send_item
    //
    // This function sends data item to AxiMM interface
    // If some package is currently trasmited, new item will be added to queue
    // Depending on ~tr_type~ in item object transaction will be posterd at proper channel
    //
    extern virtual function void send_item(item_t item);

    // Function: send_aw
    //
    // This function sends data item at AxiMM AW channel
    // If some package is currently trasmited, new item will be added to queue
    //
    extern virtual function void send_aw(item_t item);

    // Function: send_w
    //
    // This function sends data item at AxiMM W channel
    // If some package is currently trasmited, new item will be added to queue
    //
    extern virtual function void send_w(item_t item);

    // Function: send_b
    //
    // This function sends data item at AxiMM B channel
    // If some package is currently trasmited, new item will be added to queue
    //
    extern virtual function void send_b(item_t item);

    // Function: send_ar
    //
    // This function sends data item at AxiMM AR channel
    // If some package is currently trasmited, new item will be added to queue
    //
    extern virtual function void send_ar(item_t item);

    // Function: send_r
    //
    // This function sends data item at AxiMM R channel
    // If some package is currently trasmited, new item will be added to queue
    //
    extern virtual function void send_r(item_t item);

    // Function: run_phase
    //
    // Main VIPs service loop - run VIPs processing (must be called by test)
    // This task spawn internal processes and return immediately
    // Spawned processes are endless unless ~gvip_config::stop_vip~ is set
    //
    extern virtual task run_phase();

    // Function: clean
    //
    // This function clean VIP
    // Brakes any transfers at interface, clear all queues and any data structures
    //
    extern virtual function void clean();

  endclass


  //////////////////////////////////
  // Core function implementation //
  //////////////////////////////////


  function core::new (string name, virtual gaximm_if #(N, I, A, U_REQ, U_DATA, U_RESP) vif,
                          virtual gclk_if vclk, gvip::type_e vip_type, gvip_config vip_cfg = null);
    this.name = name;
    this.vif = vif;
    this.vclk = vclk;
    this.vip_type = vip_type;
    this.cfg = vip_cfg;

    // user did not pass configuration, create default one
    if (cfg == null) cfg = new();

    // depending on VIP type create master or slave handshake object
    if (vip_type == gvip::MASTER) begin
      ch_aw = aximm_aw_transmitter#(N, I, A, U_REQ, U_DATA, U_RESP, item_t)::create(vif, vclk);
      ch_ar = aximm_ar_transmitter#(N, I, A, U_REQ, U_DATA, U_RESP, item_t)::create(vif, vclk);
      ch_w  = aximm_w_transmitter#(N, I, A, U_REQ, U_DATA, U_RESP, item_t)::create(vif, vclk);
      ch_b  = aximm_b_receiver#(N, I, A, U_REQ, U_DATA, U_RESP)::create(vif, vclk);
      ch_r  = aximm_r_receiver#(N, I, A, U_REQ, U_DATA, U_RESP)::create(vif, vclk);
    end else if (vip_type == gvip::SLAVE) begin
      ch_aw = aximm_aw_receiver#(N, I, A, U_REQ, U_DATA, U_RESP)::create(vif, vclk);
      ch_ar = aximm_ar_receiver#(N, I, A, U_REQ, U_DATA, U_RESP)::create(vif, vclk);
      ch_w  = aximm_w_receiver#(N, I, A, U_REQ, U_DATA, U_RESP)::create(vif, vclk);
      ch_b  = aximm_b_transmitter#(N, I, A, U_REQ, U_DATA, U_RESP, item_t)::create(vif, vclk);
      ch_r  = aximm_r_transmitter#(N, I, A, U_REQ, U_DATA, U_RESP, item_t)::create(vif, vclk);
    end

    // create monitor handshake object
    mon_ch_aw = new(vif, vclk, ch_b);
    mon_ch_ar = new(vif, vclk, ch_r);
    mon_ch_w  = new(vif, vclk, ch_b);
    mon_ch_b  = new(vif, vclk);
    mon_ch_r  = new(vif, vclk);
    mon_aw_w  = new(mon_ch_aw, mon_ch_w);

    greporter::info(cfg.report_id, $sformatf("VIP '%s' CREATED", name));
  endfunction : new


  function gvip::type_e core::get_vip_type();
    return vip_type;
  endfunction : get_vip_type


  function gvip_config core::get_config();
    return cfg;
  endfunction : get_config


  function void core::add_ap2pkt(gvip::channel_e channel, analysis_port_ic #(item_t) ap);
    gvip_amba_axi4_pkg::analysis_port_wrp#(item_t, analysis_port_ic #(item_t)) wrp = new(ap);
    case (channel)
      gvip::AW   : mon_ch_aw.add_ap2pkt(wrp);
      gvip::AR   : mon_ch_ar.add_ap2pkt(wrp);
      gvip::W    : mon_ch_w.add_ap2pkt(wrp);
      gvip::B    : mon_ch_b.add_ap2pkt(wrp);
      gvip::R    : mon_ch_r.add_ap2pkt(wrp);
      gvip::AW_W : mon_aw_w.add_ap2pkt(wrp);
    endcase
  endfunction : add_ap2pkt


  function void core::add_ap2transfer(gvip::channel_e channel, analysis_port_ic #(item_t) ap);
    gvip_amba_axi4_pkg::analysis_port_wrp#(item_t, analysis_port_ic #(item_t)) wrp = new(ap);
    case (channel)
      gvip::AW   : mon_ch_aw.add_ap2transfer(wrp);
      gvip::AR   : mon_ch_ar.add_ap2transfer(wrp);
      gvip::W    : mon_ch_w.add_ap2transfer(wrp);
      gvip::B    : mon_ch_b.add_ap2transfer(wrp);
      gvip::R    : mon_ch_r.add_ap2transfer(wrp);
      gvip::AW_W : mon_aw_w.add_ap2pkt(wrp);
    endcase
  endfunction : add_ap2transfer


  function void core::set_policy(gvip::channel_e channel, policy obj,
                                 bit do_randomize_policy = 0, bit clone = 1);

    if (do_randomize_policy && !obj.randomize()) begin
      greporter::error(cfg.report_id, $sformatf("Policy '%s' randomizastion issue", obj.name));
    end
    greporter::info(cfg.report_id, $sformatf("Policy set to %s [%s]", obj.get_name(),
                                                                      obj.get_type_name()));
    case (channel)
      gvip::AW   : ch_aw.set_policy(obj, clone);
      gvip::AR   : ch_ar.set_policy(obj, clone);
      gvip::W    : ch_w.set_policy(obj, clone);
      gvip::B    : ch_b.set_policy(obj, clone);
      gvip::R    : ch_r.set_policy(obj, clone);
      gvip::AW_W : begin
                    ch_aw.set_policy(obj, .clone(1));
                    ch_w.set_policy(obj, .clone(1));
                   end
    endcase
  endfunction : set_policy


  function void core::send_item(item_t item);
    case (item.tr_type)
      gvip::AW  : send_aw(item);
      gvip::AR  : send_ar(item);
      gvip::W   : send_w(item);
      gvip::B   : send_b(item);
      gvip::R   : send_r(item);
      gvip::AW_W: begin
                    send_aw(item);
                    send_w(item);
                  end
    endcase
  endfunction : send_item


  function void core::send_aw(item_t item);
    channel_send #(aximm_aw_transmitter #(N, I, A, U_REQ, U_DATA, U_RESP),
                  item_t)::put(ch_aw, item, cfg, "AW");
  endfunction : send_aw


  function void core::send_w(item_t item);
    channel_send #(aximm_w_transmitter #(N, I, A, U_REQ, U_DATA, U_RESP),
                  item_t)::put(ch_w, item, cfg, "W");
  endfunction : send_w


  function void core::send_b(item_t item);
    channel_send #(aximm_b_transmitter #(N, I, A, U_REQ, U_DATA, U_RESP),
                  item_t)::put(ch_b, item, cfg, "B");
  endfunction : send_b


  function void core::send_ar(item_t item);
    channel_send #(aximm_ar_transmitter #(N, I, A, U_REQ, U_DATA, U_RESP),
                  item_t)::put(ch_ar, item, cfg, "AR");
  endfunction : send_ar


  function void core::send_r(item_t item);
    channel_send #(aximm_r_transmitter #(N, I, A, U_REQ, U_DATA, U_RESP),
                  item_t)::put(ch_r, item, cfg, "R");
  endfunction : send_r


  task core::run_phase();
    // enable interface access
    case (vip_type)
      gvip::MASTER : vif.g_sim.force_as_manager();
      gvip::SLAVE  : vif.g_sim.force_as_subordinate();
      default : begin end
    endcase

    // start core threads
    fork
      while(!cfg.stop_vip) if (ch_aw != null) ch_aw.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) if (ch_ar != null) ch_ar.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) if (ch_w != null) ch_w.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) if (ch_b != null) ch_b.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) if (ch_r != null) ch_r.start(cfg.transaction_in_reset);

      while(!cfg.stop_vip) mon_ch_aw.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) mon_ch_ar.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) mon_ch_w.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) mon_ch_b.start(cfg.transaction_in_reset);
      while(!cfg.stop_vip) mon_ch_r.start(cfg.transaction_in_reset);
    join_none
  endtask : run_phase


  function void core::clean();
    ch_aw.clean();
    ch_ar.clean();
    ch_w.clean();
    ch_b.clean();
    ch_r.clean();
    mon_ch_aw.clean();
    mon_ch_ar.clean();
    mon_ch_w.clean();
    mon_ch_b.clean();
    mon_ch_r.clean();
  endfunction : clean

endpackage: gvip_aximm_pkg
